/* Copyright 2007-2009 ENSEIRB, INRIA & CNRS
**
** This file is part of the Scotch software package for static mapping,
** graph partitioning and sparse matrix ordering.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
**
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
**
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
**
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
/************************************************************/
/**                                                        **/
/**   NAME       : hdgraph_fold.h                          **/
/**                                                        **/
/**   AUTHOR     : Francois PELLEGRINI                     **/
/**                                                        **/
/**   FUNCTION   : These lines are the data declarations   **/
/**                for the distributed source graph        **/
/**                folding routines.                       **/
/**                                                        **/
/**                # Version 5.0  : from : 06 sep 2006     **/
/**                                 to   : 06 sep 2006     **/
/**                # Version 5.1  : from : 03 jan 2009     **/
/**                                 to   : 03 jan 2009     **/
/**                                                        **/
/************************************************************/

/*
**  The defines.
*/

/* Slot indices used for point-to-point folding
   communications. First indices are used for
   communications without further processing.
   At the moment, there is one anonymous slot:
   vnumloctab.                                  */

typedef enum HdgraphFoldTag_ {
  HDGRAPHFOLDTAGENBR = 1,                         /*+ Edge size message  +*/
  HDGRAPHFOLDTAGVERT,                             /*+ vertloctab message +*/
  HDGRAPHFOLDTAGVEND,                             /*+ vendloctab message +*/
  HDGRAPHFOLDTAGVELO,                             /*+ veloloctab message +*/
  HDGRAPHFOLDTAGEDGE,                             /*+ edgeloctab message +*/
  HDGRAPHFOLDTAGNBR                               /*+ Number of tags     +*/
} HdgraphFoldTag;
