/* Copyright 2007,2021 IPB, Universite de Bordeaux, INRIA & CNRS
**
** This file is part of the Scotch software package for static mapping,
** graph partitioning and sparse matrix ordering.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
**
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
**
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
**
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
/************************************************************/
/**                                                        **/
/**   NAME       : dgraph_ghst.h                           **/
/**                                                        **/
/**   AUTHOR     : Francois PELLEGRINI                     **/
/**                Francois CHATENET (P0.0)                **/
/**                Sebastien FOUCAULT (P0.0)               **/
/**                Nicolas GICQUEL (P0.1)                  **/
/**                Jerome LACOSTE (P0.1)                   **/
/**                                                        **/
/**   FUNCTION   : These lines are the data declarations   **/
/**                for the halo building routine.          **/
/**                                                        **/
/**                # Version P0.0 : from : 01 apr 1997     **/
/**                                 to   : 20 jun 1997     **/
/**                # Version P0.1 : from : 12 apr 1998     **/
/**                                 to   : 20 jun 1998     **/
/**                # Version 5.0  : from : 28 feb 2006     **/
/**                                 to   : 31 dec 2006     **/
/**                # Version 6.1  : from : 04 dec 2021     **/
/**                                 to   : 04 dec 2021     **/
/**                # Version 7.0  : from : 22 mar 2023     **/
/**                                 to   : 22 mar 2023     **/
/**                                                        **/
/************************************************************/

/*
** The defines.
*/

/* procsidtab-related values. */

#define DGRAPHGHSTSIDMAX            ((int) ((1U << (sizeof (unsigned int) * 8 - 1)) - 2U)) /* Maximum leap value for procsidtab entries */

/*
** The type and structure definitions.
*/

/* Sort structure for ghost edges. */

typedef struct DgraphGhstSort_ {
  Gnum              vertglbnum;                   /* Global end vertex number */
  Gnum              edgegstnum;                   /* Number of ghost edge     */
} DgraphGhstSort;
