/* Copyright 2017-2019,2023 IPB, Universite de Bordeaux, INRIA & CNRS
**
** This file is part of the Scotch software package for static mapping,
** graph partitioning and sparse matrix ordering.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
**
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
**
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
**
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
/************************************************************/
/**                                                        **/
/**   NAME       : library_graph_diam.c                    **/
/**                                                        **/
/**   AUTHOR     : Francois PELLEGRINI                     **/
/**                                                        **/
/**   FUNCTION   : This module is the API for the source   **/
/**                graph handling routines of the          **/
/**                libSCOTCH library.                      **/
/**                                                        **/
/**   DATES      : # Version 6.0  : from : 26 jan 2017     **/
/**                                 to   : 22 apr 2018     **/
/**                # Version 7.0  : from : 07 may 2019     **/
/**                                 to   : 21 jan 2023     **/
/**                                                        **/
/************************************************************/

/*
**  The defines and includes.
*/

#include "module.h"
#include "common.h"
#include "context.h"
#include "graph.h"
#include "scotch.h"

/************************************/
/*                                  */
/* These routines are the C API for */
/* the graph handling routines.     */
/*                                  */
/************************************/

/*+ This routine returns the vertex-traversal
*** pseudo-diameter of the given graph.
*** It returns:
*** - [0; GNUMMAX[ : graph pseudo-diameter.
*** - GNUMMAX      : disconnected graph.
*** - < 0          : on error.
+*/

SCOTCH_Num
SCOTCH_graphDiamPV (
const SCOTCH_Graph * const  libgrafptr)
{
  CONTEXTDECL        (libgrafptr);
  SCOTCH_Num          diamval;

  if (CONTEXTINIT (libgrafptr) != 0) {
    errorPrint (STRINGIFY (SCOTCH_graphDiamPV) ": cannot initialize context");
    return     (1);
  }

  diamval = (SCOTCH_Num) graphDiamPV ((Graph * const) CONTEXTGETOBJECT (libgrafptr), CONTEXTGETDATA (libgrafptr));

  CONTEXTEXIT (libgrafptr);
  return      (diamval);
}
