/* Copyright 2004,2007,2023 IPB, Universite de Bordeaux, INRIA & CNRS
**
** This file is part of the Scotch software package for static mapping,
** graph partitioning and sparse matrix ordering.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
**
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
**
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
**
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
/************************************************************/
/**                                                        **/
/**   NAME       : geom.c                                  **/
/**                                                        **/
/**   AUTHOR     : Francois PELLEGRINI                     **/
/**                                                        **/
/**   FUNCTION   : This module handles the geometric       **/
/**                source graph functions.                 **/
/**                                                        **/
/**   DATES      : # Version 3.3  : from : 12 dec 1998     **/
/**                                 to   : 21 dec 1998     **/
/**                # Version 4.0  : from : 18 dec 2001     **/
/**                                 to   : 26 nov 2003     **/
/**                # Version 7.0  : from : 18 jan 2023     **/
/**                                 to   : 18 jan 2023     **/
/**                                                        **/
/************************************************************/

/*
**  The defines and includes.
*/

#include "module.h"
#include "common.h"
#include "geom.h"

/********************************************/
/*                                          */
/* These routines handle geometrical gdata. */
/*                                          */
/********************************************/

/* This routine initializes a geometrical
** data structure.
** It returns:
** - 0  : in all cases.
*/

int
geomInit (
Geom * restrict const       geomptr)
{
  geomptr->dimnnbr = 0;                           /* Initialize geometry */
  geomptr->geomtab = NULL;

  return (0);
}

/* This routine frees a geometrical graph structure.
** It returns:
** - VOID  : in all cases.
*/

void
geomExit (
Geom * restrict const       geomptr)
{
  if (geomptr->geomtab != NULL)
    memFree (geomptr->geomtab);

  geomptr->dimnnbr = 0;
  geomptr->geomtab = NULL;
}
