/* Copyright 2018,2019,2021 IPB, Universite de Bordeaux, INRIA & CNRS
**
** This file is part of the Scotch software package for static mapping,
** graph partitioning and sparse matrix ordering.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
**
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
**
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
**
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
/************************************************************/
/**                                                        **/
/**   NAME       : context.h                               **/
/**                                                        **/
/**   AUTHOR     : Francois PELLEGRINI                     **/
/**                                                        **/
/**   FUNCTION   : These lines are the data declarations   **/
/**                for the execution context handling      **/
/**                routines.                               **/
/**                                                        **/
/**   DATES      : # Version 7.0  : from : 04 aug 2018     **/
/**                                 to   : 08 oct 2021     **/
/**                                                        **/
/************************************************************/

#define CONTEXT_H

/*
**  The defines.
*/

/*+ Container option flags. +*/

#define CONTEXTCONTAINERTYPE        0x4000        /*+ Object is a container +*/

/*+ The container integer option values. +*/

typedef enum {
  CONTEXTOPTIONNUMDETERMINISTIC,
  CONTEXTOPTIONNUMRANDOMFIXEDSEED,
  CONTEXTOPTIONNUMNBR
} ContextOptionNum;

/*+ The container double option values. +*/

typedef enum {
  CONTEXTOPTIONDBLNBR
} ContextOptionDbl;

/*
**  The type and structure definitions.
*/

/*+ The container flag type. +*/

typedef unsigned int ContextContainerFlag;        /*+ Container flag; TRICK: same type as first field of contained objects +*/

/*+ The context container structure. It
    simulates the shape of objects to which
    a context is attached.                  +*/

typedef struct ContextContainer_ {
  ContextContainerFlag      flagval;              /*+ TRICK: flag area similar to that of attached objects +*/
  Context *                 contptr;              /*+ Context attached to the object                       +*/
  void *                    dataptr;              /*+ Object to which the context is attached              +*/
} ContextContainer;

/*
**  The function prototypes.
*/

int                         contextOptionsInit  (Context * const);

/*
**  The macro definitions.
*/

#define contextContainerTrue(c)     ((*((ContextContainerFlag *) (c)) & CONTEXTCONTAINERTYPE) != 0)

#define contextContainerData(c)     (((ContextContainer *) (c))->contptr)
#define contextContainerObject(c)   (((ContextContainer *) (c))->dataptr)

#define CONTEXTDECL(c)             \
  Context *           c##_contptr; \
  Context             c##_contdat; \
  void *              c##_cdatptr

#define CONTEXTINIT(c)                          \
  ((contextContainerTrue (c))                   \
   ? (c##_contptr = contextContainerData   (c), \
      c##_cdatptr = contextContainerObject (c), \
      0)                                        \
   : (c##_contptr = (void *) &c##_contdat,      \
      c##_cdatptr = (void *) (c),               \
      contextInit        (&c##_contdat),        \
      contextOptionsInit (&c##_contdat),        \
      contextCommit      (&c##_contdat)))

#define CONTEXTGETDATA(c)           c##_contptr
#define CONTEXTGETOBJECT(c)         c##_cdatptr

#define CONTEXTEXIT(c)             \
  if (c##_contptr == &c##_contdat) \
    contextExit (&c##_contdat)

#define CONTEXTOBJECT(c)        \
  ((contextContainerTrue (c))   \
   ? contextContainerObject (c) \
   : ((void *) (c)))
